#!/usr/bin/env --split-string=python -m pytest --verbose

"""Self test code"""

import pytest

from wanip import Providers, Public_Providers

class Testcase_Providers_01:

    def test_generals(self):

        global Public_Providers

        assert len(Public_Providers) > 1
        assert "my machine" not in Public_Providers
        assert repr(Public_Providers) is not None
        assert str(Public_Providers) is not None
        P = Public_Providers()
        assert P in Public_Providers

        assert "https://ifconfig.me/ip" in Public_Providers

        Public_Providers = Providers(1, 2, 3, 4, 5, 6)
        assert len(Public_Providers) == 6
        assert 6 in Public_Providers

        Public_Providers = Public_Providers + 42
        assert len(Public_Providers) == 7
        assert 42 in Public_Providers

        Public_Providers = Public_Providers + 'foo'
        assert len(Public_Providers) == 8
        assert 'foo' in Public_Providers

        Public_Providers = Public_Providers + "bar"
        assert len(Public_Providers) == 9
        assert "baz" not in Public_Providers

        Public_Providers = Public_Providers + "zonk" + 'baz'
        assert len(Public_Providers) == 11
        assert "baz" in Public_Providers

        Public_Providers = Public_Providers - "zonk" - 3
        assert len(Public_Providers) == 9
        assert 3 not in Public_Providers

        Public_Providers = Public_Providers - 4
        assert len(Public_Providers) == 8
        assert 4 not in Public_Providers

class Testcase_Providers_02:

    def test_removing_an_unknown_element(self):
        # Cannot remove non-existing element
        with Providers() as P:
            P = P + 'a known element'
            with pytest.raises(ValueError):
                P = P - 'non-existing element'

    def test_adding_things_twice(self):
        # Cannot add an element more than once
        with Providers() as P:
            with pytest.raises(ValueError):
                P = P + 'an element' + 'an element'

class Testcase_Providers_03:

    def test_chache(self):

        P = Providers(1, 2, 3, 4)
        a = P()
        assert a in P

        # Make sure a cannot be returned again
        P = P - a
        assert a not in P

        b = P()
        assert a != b

        # Put a back into P
        P = P + a
        assert a in P

        a, b = P(use_cached=True), P(use_cached=not False)
        assert a == b

        # Now have cached providers to begin with
        Q = Providers(4711, 3206, 42, use_cached=True)
        c = Q()
        assert c in Q

        Q = Q - c
        assert c not in Q

        d = Q(use_cached=True)
        assert d == c

        d = Q()
        assert d != c

class Testcase_Providers_04:

    def test_providers_drainage(self):
        # Drain providers
        P = Providers(self.test_providers_drainage.__name__)
        P = P - P()
        with pytest.raises(ValueError):
            P = P - P()

    def test_telescoping_series(self):
        P = Providers()
        P = P + 1
        P = P + 2 - 2 + 2 - 2 + 2 - 2 + 'a' + 'b' - 'a' - 'b'
        P = P + 9
        Q = Providers(1, 9)

        assert len(P) == 2
        assert len(Q) == 2

        p1 = P(); P = P - p1
        p2 = P(); P = P - p2

        q1 = Q(); Q = Q - q1
        q2 = Q(); Q = Q - q2

        assert sorted((p1, p2)) == sorted((q1, q2))

    def test_empty_list_of_providers(self):

        # Cannot obtain a provider from an empty collection
        Public_Providers = Providers()
        with pytest.raises(ValueError):
            print(Public_Providers())

        # Not even when asking for a cached one
        with pytest.raises(ValueError):
            print(Public_Providers(use_cached=not False))

class Testcase_Providers_05:

    def test_context_manager_protocol(self):
        # Test the context manager protocol
        with Providers('spam', 'eggs') as P:
            P()
        with pytest.raises(ValueError):
            # We have left the context (of Providers(...), that is)
            P()



"""Ask a provider for your ip with which you connect to it, then print it out"""

import subprocess
import sys
import argparse
from pathlib import PurePath
from typing import Optional

from .providers import Public_Providers, Providers

class Wanip:

    def __init__(self,
                 me: Optional[str] = PurePath(__file__).stem,
                 purpose : Optional[str] = __doc__) -> None:
        """Kick off scanning the command-line"""
        self.args = self.parse_cmd_line(me, purpose)
        self.args.ipv4 = '--ipv4' if self.args.ipv4 else str()
        self.args.ipv6 = '--ipv6' if self.args.ipv6 else str()

    def curlme(self, provider: str) -> None:
        """Use curl to get hold of my WANIP"""
        # Construct curl command; note that -4 and -6 cannot be used together but since the
        # ArgumentParser takes care of at most one having survived in the args, their
        # representations "can" in the f-string.
        curl_cmd = f'curl {self.args.ipv4} {self.args.ipv6} --fail --show-error --silent {provider}'

        # Obtain data & report
        result = subprocess.run(curl_cmd, shell=True, capture_output=True)
        if result.returncode != 0:
            print(f'Cannot retrieve your WAN IP address from "{provider}"', file=sys.stderr)
            print(f'The error message is this: {result.stderr.decode()}', end='', file=sys.stderr)
        else:
            print('{output}'.format(output=result.stdout.decode().rstrip()))

    def parse_cmd_line(self, me: str, purpose: str) -> Optional[argparse.Namespace]:
        """Read options, show help"""
        # Parse the command line
        try:
            parser = argparse.ArgumentParser(
                prog=me,
                description=purpose,
                epilog='Respect the netiquette when contacting the provider.',
                formatter_class=argparse.ArgumentDefaultsHelpFormatter,
            )
            parser.add_argument(
                '-p', '--provider',
                type=str,
                help='''
                    the provider to contact, instead of pseudo-randomly auto-selecting one from a
                    pre-built internal list
                    ''',
            )
            group = parser.add_mutually_exclusive_group()
            group.add_argument(
                '-4', '--ipv4',
                action='store_true',
                help='''
                    force the usage of IPv4
                    ''',
            )
            group.add_argument(
                '-6', '--ipv6',
                action='store_true',
                help='''
                    force the usage of IPv6
                    ''',
            )
            parser.add_argument(
                '-v', '--verbose',
                action='count',
                default=0,
                help='''
                    used once: show which provider will be contacted; used twice (or more often):
                    display contactable providers as well (i.e., the pre-built internal list)
                    ''',
            )
            return parser.parse_args()
        except argparse.ArgumentError as exc:
            raise ValueError('The command-line is indecipherable')

    def __call__(self) -> None:
        """Run the show

        The LEGB scoping rule means that in order to overwrite the Public_Providers (a sensible
        thing to do in case the -p/--provider option has been specified), we must declare it as
        global.
        """
        global Public_Providers

        if self.args.provider:
            Public_Providers = Providers(self.args.provider)
        self.args.provider = Public_Providers()

        if self.args.verbose > 1: print(Public_Providers)
        if self.args.verbose >= 1: print('Trying {provider}'.format(provider=self.args.provider))

        self.curlme(self.args.provider)

def __main():

    W = Wanip()
    W()

def main():
    try:
        __main()
    except Exception:
        import traceback
        print(traceback.format_exc(), file=sys.stderr, end='')
        sys.exit(1)
    except KeyboardInterrupt:
        print('Interrupted by user', file=sys.stderr, end='')
        sys.exit(1)

if __name__ == '__main__':
    main()



"""Wrap 'public providers for WANIP' in a data structure with convenience operations

It is really our interpretation that those be labelled thusly...
"""

import random
from typing import Iterator, Any, Self

from lsattr import LsAttr

__all__ = [
    'Public_Providers',
    'Providers',
]

class Providers(LsAttr):

    def __init__(self, *providers, use_cached=False, cached=None):
        """Store the providers in a tuple

        This feels more natural than, say, a list or a set; although we will resort to the former
        and emulate semantix of the latter below.
        """
        self.providers = tuple(providers)
        self.use_cached = use_cached
        self.cached = cached

    def __call__(self, use_cached=False):
        """Gimme any one"""

        if not use_cached:
            try:
                assert len(self.providers) >= 1
            # We catch TypeError in case we have left the context: len(None)...
            except (AssertionError, TypeError) as exc:
                raise ValueError('no element found in collection') from exc

            self.cached = random.choice(self.providers)
            return self.cached
        else:
            try:
                assert self.cached is not None
                return self.cached
            except AssertionError as exc:
                raise ValueError('no element found in cache') from exc

    def __add__(self, provider):
        """Augment the collection

        As tuples are immutable, we temporarily escape to lists.  We do not allow an element to be
        added more than once.
        """
        try:
            assert provider not in self.providers
        except AssertionError as exc:
            raise ValueError(f'"{provider}" is already part of the collection') from exc

        _l = list(self.providers)
        _l.append(provider)
        return self.__class__(*(_p for _p in _l), use_cached=self.use_cached, cached=self.cached)

    def __sub__(self, provider):
        """Diminish the collection

        Again, we resort to lists while doing the job.
        """
        _l = list(self.providers)

        try:
            _l.remove(provider)
        except ValueError as exc:
            raise ValueError(f'"{provider}" is not part of the collection') from exc

        return self.__class__(*(_p for _p in _l), use_cached=self.use_cached, cached=self.cached)

    def __len__(self):
        """Tell the size of the collection"""

        return len(self.providers)

    def __contains__(self, provider):
        """Test for membership

        We waive the "in" operator so that we can re-use the iterator.
        """
        for _candidate in self:
            if _candidate == provider:
                return True
        return False

    def __iter__(self) -> Iterator[Any]:
        """Make the collection iterable/usable in generators"""

        for _p in self.providers:
            yield _p

    def __str__(self):
        """Pretty-print"""

        return '\n'.join(sorted(str(provider) for provider in self))

    def __enter__(self) -> Self:
        """Enter context"""

        return self

    def __exit__(self, exc_value, exc_type, traceback):
        """Leave context"""

        self.providers = None
        self.cached    = None
        return False

# Export this
Public_Providers = Providers(
    'https://ifconfig.me/ip',
    'https://my.ip.fi',
    'https://icanhazip.com',
    'https://ifconfig.co',
    'https://ipecho.net/plain',
    'https://ipinfo.io/ip',
    'https://ident.me',
    'https://iprs.fly.dev',
    'https://l2.io/ip',
    'https://ipapi.co/ip',
    'https://wgetip.com',
    'https://whatismyip.akamai.com',
    'https://eth0.me/',
    'https://api.ipify.org',
    'https://ip.me',
    'https://checkip.amazonaws.com',
    'https://www.trackip.net/ip',
    'https://corz.org/ip',
)
